@@ -91,6 +91,9 @@ struct CompilationOptions {
     /// giving up due to potentially cyclic dependencies in parameter values.
     uint32_t maxDefParamSteps = 128;
 
+    /// The maximum number of instances allowed in a single instance array.
+    uint32_t maxInstanceArray = 65535;
+
     /// The maximum number of errors that can be found before we short circuit
     /// the tree walking process.
     uint32_t errorLimit = 64;
@@ -192,8 +195,8 @@ public:
     const Definition* getDefinition(const ModuleDeclarationSyntax& syntax) const;
 
     /// Creates a new definition in the given scope based on the given syntax.
-    const Definition& createDefinition(const Scope& scope, LookupLocation location,
-                                       const ModuleDeclarationSyntax& syntax);
+    void createDefinition(const Scope& scope, LookupLocation location,
+                          const ModuleDeclarationSyntax& syntax);
 
     /// Gets the package with the give name, or null if there is no such package.
     const PackageSymbol* getPackage(string_view name) const;
@@ -219,9 +222,17 @@ public:
     /// Registers a system subroutine handler, which can be accessed by compiled code.
     void addSystemSubroutine(std::unique_ptr<SystemSubroutine> subroutine);
 
+    /// Registers an externally owned system subroutine handler,
+    /// which can be accessed by compiled code.
+    void addSystemSubroutine(const SystemSubroutine& subroutine);
+
     /// Registers a type-based system method handler, which can be accessed by compiled code.
     void addSystemMethod(SymbolKind typeKind, std::unique_ptr<SystemSubroutine> method);
 
+    /// Registers an externally owned type-based system method handler,
+    /// which can be accessed by compiled code.
+    void addSystemMethod(SymbolKind typeKind, const SystemSubroutine& subroutine);
+
     /// Gets a system subroutine with the given name, or null if there is no such subroutine
     /// registered.
     const SystemSubroutine* getSystemSubroutine(string_view name) const;
@@ -313,6 +324,10 @@ public:
     /// Notes that there is a default disable associated with the specified scope.
     void noteDefaultDisable(const Scope& scope, const Expression& expr);
 
+    /// Notes that the given symbol has a name conflict in its parent scope.
+    /// This will cause appropriate errors to be issued.
+    void noteNameConflict(const Symbol& symbol);
+
     /// Finds an applicable default disable expression for the given scope, or returns nullptr
     /// if no such declaration is in effect.
     const Expression* getDefaultDisable(const Scope& scope) const;
@@ -493,10 +508,10 @@ private:
     flat_hash_map<string_view, const PackageSymbol*> packageMap;
 
     // The name map for system subroutines.
-    flat_hash_map<string_view, std::unique_ptr<SystemSubroutine>> subroutineMap;
+    flat_hash_map<string_view, const SystemSubroutine*> subroutineMap;
 
     // The name map for system methods.
-    flat_hash_map<std::tuple<string_view, SymbolKind>, std::unique_ptr<SystemSubroutine>> methodMap;
+    flat_hash_map<std::tuple<string_view, SymbolKind>, const SystemSubroutine*> methodMap;
 
     // Map from pointers (to symbols, statements, expressions) to their associated attributes.
     flat_hash_map<const void*, span<const AttributeSymbol* const>> attributeMap;
@@ -594,6 +609,9 @@ private:
     // A list of extern interface method implementations for later elaboration.
     std::vector<const SubroutineSymbol*> externInterfaceMethods;
 
+    // A list of name conflicts to later resolve by issuing diagnostics.
+    std::vector<const Symbol*> nameConflicts;
+
     // A map of scopes to default clocking blocks.
     flat_hash_map<const Scope*, const Symbol*> defaultClockingMap;
 
@@ -603,6 +621,9 @@ private:
     // A map of scopes to default disable declarations.
     flat_hash_map<const Scope*, const Expression*> defaultDisableMap;
 
+    // Storage for system subroutine instances.
+    std::vector<std::unique_ptr<SystemSubroutine>> subroutineStorage;
+
     // The built-in std package.
     const PackageSymbol* stdPkg = nullptr;
 };
