@@ -797,7 +797,7 @@ const Symbol* findSuperHandle(const Scope& scope, SourceRange range, LookupResul
     }
 
     auto base = parent->getBaseClass();
-    if (!base) {
+    if (!base && !parent->name.empty()) {
         result.addDiag(scope, diag::SuperNoBase, range) << parent->name;
         return nullptr;
     }
@@ -872,6 +872,24 @@ void Lookup::name(const NameSyntax& syntax, const BindContext& context, bitmask<
             }
             result.found = &scope.getCompilation().getRoot();
             return;
+        case SyntaxKind::ConstructorName:
+            result.addDiag(scope, diag::UnexpectedNameToken, syntax.sourceRange())
+                << syntax.getFirstToken().valueText();
+            result.found = nullptr;
+            return;
+        case SyntaxKind::LocalScope:
+            // This can only happen in error scenarios, where the parser has
+            // already issued a diagnostic.
+            result.found = nullptr;
+            return;
+        case SyntaxKind::UnitScope:
+        case SyntaxKind::SuperHandle:
+        case SyntaxKind::ArrayUniqueMethod:
+        case SyntaxKind::ArrayAndMethod:
+        case SyntaxKind::ArrayOrMethod:
+        case SyntaxKind::ArrayXorMethod:
+            // These error cases can't happen here because the parser will always
+            // wrap them into a scoped name.
         default:
             THROW_UNREACHABLE;
     }
@@ -1675,8 +1693,6 @@ void Lookup::qualified(const ScopedNameSyntax& syntax, const BindContext& contex
     SyntaxKind firstKind = leftMost->kind;
     NameComponents first = *leftMost;
     auto name = first.text;
-    if (name.empty())
-        return;
 
     auto popFront = [&] {
         first = nameParts.back().name;
@@ -1703,6 +1719,9 @@ void Lookup::qualified(const ScopedNameSyntax& syntax, const BindContext& contex
         popFront();
     }
 
+    if (name.empty())
+        return;
+
     switch (firstKind) {
         case SyntaxKind::IdentifierName:
         case SyntaxKind::IdentifierSelectName:
@@ -1713,6 +1732,9 @@ void Lookup::qualified(const ScopedNameSyntax& syntax, const BindContext& contex
         case SyntaxKind::UnitScope: {
             // Walk upward to find the compilation unit scope.
             popFront();
+            if (name.empty())
+                return;
+
             const Scope* current = &scope;
             LookupLocation location = context.getLocation();
             do {
@@ -1755,6 +1777,13 @@ void Lookup::qualified(const ScopedNameSyntax& syntax, const BindContext& contex
             // This is only reachable in invalid code. An error has already been
             // reported so early out.
             return;
+        case SyntaxKind::ArrayUniqueMethod:
+        case SyntaxKind::ArrayAndMethod:
+        case SyntaxKind::ArrayOrMethod:
+        case SyntaxKind::ArrayXorMethod:
+        case SyntaxKind::ConstructorName:
+            result.addDiag(scope, diag::UnexpectedNameToken, first.range) << name;
+            return;
         default:
             THROW_UNREACHABLE;
     }
