@@ -6,7 +6,6 @@
 //------------------------------------------------------------------------------
 #include "slang/compilation/Compilation.h"
 
-#include "../text/CharInfo.h"
 #include "ElabVisitors.h"
 
 #include "slang/binding/SystemSubroutine.h"
@@ -18,6 +17,7 @@
 #include "slang/parsing/Parser.h"
 #include "slang/parsing/Preprocessor.h"
 #include "slang/syntax/SyntaxTree.h"
+#include "slang/text/CharInfo.h"
 #include "slang/text/SourceManager.h"
 #include "slang/types/TypePrinter.h"
 
@@ -441,13 +441,15 @@ const Definition* Compilation::getDefinition(const ModuleDeclarationSyntax& synt
 
 template<typename T, typename U>
 static void reportRedefinition(const Scope& scope, const T& newSym, const U& oldSym) {
-    auto& diag = scope.addDiag(diag::Redefinition, newSym.location);
-    diag << newSym.name;
-    diag.addNote(diag::NotePreviousDefinition, oldSym.location);
+    if (!newSym.name.empty()) {
+        auto& diag = scope.addDiag(diag::Redefinition, newSym.location);
+        diag << newSym.name;
+        diag.addNote(diag::NotePreviousDefinition, oldSym.location);
+    }
 }
 
-const Definition& Compilation::createDefinition(const Scope& scope, LookupLocation location,
-                                                const ModuleDeclarationSyntax& syntax) {
+void Compilation::createDefinition(const Scope& scope, LookupLocation location,
+                                   const ModuleDeclarationSyntax& syntax) {
     auto& metadata = definitionMetadata[&syntax];
     auto def =
         std::make_unique<Definition>(scope, location, syntax, *metadata.defaultNetType,
@@ -456,25 +458,28 @@ const Definition& Compilation::createDefinition(const Scope& scope, LookupLocati
     // Record that the given scope contains this definition. If the scope is a compilation unit, add
     // it to the root scope instead so that lookups from other compilation units will find it.
     auto targetScope = scope.asSymbol().kind == SymbolKind::CompilationUnit ? root.get() : &scope;
-    auto it = definitionMap.emplace(std::tuple(def->name, targetScope), std::move(def)).first;
+    std::tuple key(def->name, targetScope);
+    if (auto it = definitionMap.find(key); it != definitionMap.end()) {
+        reportRedefinition(scope, *def, *it->second);
+        return;
+    }
+
+    auto [it, inserted] = definitionMap.emplace(key, std::move(def));
+    ASSERT(inserted);
 
     auto result = it->second.get();
     if (targetScope == root.get()) {
         auto& topDef = topDefinitions[result->name].first;
-        if (topDef)
-            reportRedefinition(scope, *result, *topDef);
-        else {
-            topDef = result;
-            if (auto primIt = udpMap.find(result->name); primIt != udpMap.end())
-                reportRedefinition(scope, *result, *primIt->second);
-        }
+        ASSERT(!topDef);
+
+        topDef = result;
+        if (auto primIt = udpMap.find(result->name); primIt != udpMap.end())
+            reportRedefinition(scope, *result, *primIt->second);
     }
     else {
         // Record the fact that we have nested modules with this given name.
         topDefinitions[result->name].second = true;
     }
-
-    return *result;
 }
 
 const PackageSymbol* Compilation::getPackage(string_view lookupName) const {
@@ -491,7 +496,7 @@ const PackageSymbol& Compilation::createPackage(const Scope& scope,
         PackageSymbol::fromSyntax(scope, syntax, *metadata.defaultNetType, metadata.timeScale);
 
     auto [it, inserted] = packageMap.emplace(package.name, &package);
-    if (!inserted) {
+    if (!inserted && !package.name.empty()) {
         auto& diag = scope.addDiag(diag::Redefinition, package.location);
         diag << package.name;
         diag.addNote(diag::NotePreviousDefinition, it->second->location);
@@ -536,25 +541,35 @@ void Compilation::addGateType(const PrimitiveSymbol& prim) {
 }
 
 void Compilation::addSystemSubroutine(std::unique_ptr<SystemSubroutine> subroutine) {
-    subroutineMap.emplace(subroutine->name, std::move(subroutine));
+    subroutineMap.emplace(subroutine->name, subroutine.get());
+    subroutineStorage.emplace_back(std::move(subroutine));
+}
+
+void Compilation::addSystemSubroutine(const SystemSubroutine& subroutine) {
+    subroutineMap.emplace(subroutine.name, &subroutine);
 }
 
 void Compilation::addSystemMethod(SymbolKind typeKind, std::unique_ptr<SystemSubroutine> method) {
-    methodMap.emplace(std::make_tuple(string_view(method->name), typeKind), std::move(method));
+    methodMap.emplace(std::make_tuple(string_view(method->name), typeKind), method.get());
+    subroutineStorage.emplace_back(std::move(method));
+}
+
+void Compilation::addSystemMethod(SymbolKind typeKind, const SystemSubroutine& method) {
+    methodMap.emplace(std::make_tuple(string_view(method.name), typeKind), &method);
 }
 
 const SystemSubroutine* Compilation::getSystemSubroutine(string_view name) const {
     auto it = subroutineMap.find(name);
     if (it == subroutineMap.end())
         return nullptr;
-    return it->second.get();
+    return it->second;
 }
 
 const SystemSubroutine* Compilation::getSystemMethod(SymbolKind typeKind, string_view name) const {
     auto it = methodMap.find(std::make_tuple(name, typeKind));
     if (it == methodMap.end())
         return nullptr;
-    return it->second.get();
+    return it->second;
 }
 
 void Compilation::setAttributes(const Symbol& symbol,
@@ -638,7 +653,7 @@ void Compilation::addOutOfBlockDecl(const Scope& scope, const ScopedNameSyntax&
     auto [it, inserted] = outOfBlockDecls.emplace(std::make_tuple(className, declName, &scope),
                                                   std::make_tuple(&syntax, &name, index, false));
 
-    if (!inserted) {
+    if (!inserted && !className.empty() && !declName.empty()) {
         std::string combined = fmt::format("{}::{}", className, declName);
         auto range = std::get<1>(it->second)->sourceRange();
 
@@ -740,6 +755,10 @@ void Compilation::noteDefaultDisable(const Scope& scope, const Expression& expr)
     }
 }
 
+void Compilation::noteNameConflict(const Symbol& symbol) {
+    nameConflicts.push_back(&symbol);
+}
+
 const Expression* Compilation::getDefaultDisable(const Scope& scope) const {
     auto curr = &scope;
     while (true) {
@@ -804,13 +823,22 @@ const Diagnostics& Compilation::getSemanticDiagnostics() {
     getRoot().visit(visitor);
     visitor.finalize();
 
+    // Note for the following checks here: anything that depends on a list
+    // stored in the compilation object should think carefully about taking
+    // a copy of that list first before iterating over it, because your check
+    // might trigger additional action that ends up adding to that list,
+    // causing undefined behavior.
+
     // Check all DPI methods for correctness.
     if (!dpiExports.empty() || !visitor.dpiImports.empty())
         checkDPIMethods(visitor.dpiImports);
 
     // Check extern interface methods for correctness.
-    for (auto method : externInterfaceMethods)
-        method->connectExternInterfacePrototype();
+    if (!externInterfaceMethods.empty()) {
+        auto methods = externInterfaceMethods;
+        for (auto method : methods)
+            method->connectExternInterfacePrototype();
+    }
 
     if (!visitor.externIfaceProtos.empty())
         checkExternIfaceMethods(visitor.externIfaceProtos);
@@ -818,24 +846,37 @@ const Diagnostics& Compilation::getSemanticDiagnostics() {
     if (!visitor.modportsWithExports.empty())
         checkModportExports(visitor.modportsWithExports);
 
-    // Report on unused out-of-block definitions. These are always a real error.
-    for (auto& [key, val] : outOfBlockDecls) {
-        auto& [syntax, name, index, used] = val;
-        if (!used) {
-            auto& [className, declName, scope] = key;
-            auto classRange = name->left->sourceRange();
-            auto sym = Lookup::unqualifiedAt(*scope, className,
-                                             LookupLocation(scope, uint32_t(index)), classRange);
+    // Report any lingering name conflicts.
+    if (!nameConflicts.empty()) {
+        auto conflicts = nameConflicts;
+        for (auto symbol : conflicts) {
+            auto scope = symbol->getParentScope();
+            ASSERT(scope);
+            scope->handleNameConflict(*symbol);
+        }
+    }
 
-            if (sym) {
-                if (sym->kind == SymbolKind::ClassType ||
-                    sym->kind == SymbolKind::GenericClassDef) {
-                    auto& diag = scope->addDiag(diag::NoDeclInClass, name->sourceRange());
-                    diag << declName << className;
-                }
-                else {
-                    auto& diag = scope->addDiag(diag::NotAClass, classRange);
-                    diag << className;
+    // Report on unused out-of-block definitions. These are always a real error.
+    if (!outOfBlockDecls.empty()) {
+        auto decls = outOfBlockDecls;
+        for (auto& [key, val] : decls) {
+            auto& [syntax, name, index, used] = val;
+            if (!used) {
+                auto& [className, declName, scope] = key;
+                auto classRange = name->left->sourceRange();
+                auto sym = Lookup::unqualifiedAt(
+                    *scope, className, LookupLocation(scope, uint32_t(index)), classRange);
+
+                if (sym && !declName.empty() && !className.empty()) {
+                    if (sym->kind == SymbolKind::ClassType ||
+                        sym->kind == SymbolKind::GenericClassDef) {
+                        auto& diag = scope->addDiag(diag::NoDeclInClass, name->sourceRange());
+                        diag << declName << className;
+                    }
+                    else {
+                        auto& diag = scope->addDiag(diag::NotAClass, classRange);
+                        diag << className;
+                    }
                 }
             }
         }
@@ -1189,7 +1230,8 @@ void Compilation::checkDPIMethods(span<const SubroutineSymbol* const> dpiImports
 
     flat_hash_map<std::tuple<string_view, const Scope*>, const DPIExportSyntax*> exportsByScope;
     flat_hash_map<const SubroutineSymbol*, const DPIExportSyntax*> previousExports;
-    for (auto [syntax, scope] : dpiExports) {
+    auto exports = dpiExports;
+    for (auto [syntax, scope] : exports) {
         if (syntax->specString.valueText() == "DPI")
             scope->addDiag(diag::DPISpecDisallowed, syntax->specString.range());
 
