@@ -6,19 +6,21 @@
 //------------------------------------------------------------------------------
 #include "slang/parsing/Lexer.h"
 
-#include "../text/CharInfo.h"
 #include <algorithm>
 #include <cmath>
 
 #include "slang/diagnostics/LexerDiags.h"
 #include "slang/diagnostics/NumericDiags.h"
 #include "slang/syntax/SyntaxNode.h"
+#include "slang/text/CharInfo.h"
 #include "slang/text/SourceManager.h"
 #include "slang/util/BumpAllocator.h"
 #include "slang/util/String.h"
 
 static_assert(std::numeric_limits<double>::is_iec559, "SystemVerilog requires IEEE 754");
 
+static const double BitsPerDecimal = log2(10.0);
+
 namespace slang {
 
 using LF = LexerFacts;
@@ -37,7 +39,7 @@ Lexer::Lexer(BufferID bufferId, string_view source, const char* startPtr, BumpAl
     ASSERT(count);
     ASSERT(sourceEnd[-1] == '\0');
 
-    // detect BOMs so we can give nice errors for invaild encoding
+    // detect BOMs so we can give nice errors for invalid encoding
     if (count >= 2) {
         const unsigned char* ubuf = reinterpret_cast<const unsigned char*>(sourceBuffer);
         if ((ubuf[0] == 0xFF && ubuf[1] == 0xFE) || (ubuf[0] == 0xFE && ubuf[1] == 0xFF)) {
@@ -46,11 +48,9 @@ Lexer::Lexer(BufferID bufferId, string_view source, const char* startPtr, BumpAl
             advance(2);
         }
         else if (count >= 3) {
-            if (ubuf[0] == 0xEF && ubuf[1] == 0xBB && ubuf[2] == 0xBF) {
-                errorCount++;
-                addDiag(diag::UnicodeBOM, 0);
+            // Silently skip the UTF8 BOM.
+            if (ubuf[0] == 0xEF && ubuf[1] == 0xBB && ubuf[2] == 0xBF)
                 advance(3);
-            }
         }
     }
 }
@@ -570,20 +570,25 @@ Token Lexer::lexToken(KeywordVersion keywordVersion) {
             return create(TokenKind::Tilde);
         default:
             errorCount++;
-            if (isASCII(c))
+            if (isASCII(c)) {
                 addDiag(diag::NonPrintableChar, currentOffset() - 1);
+            }
             else {
-                addDiag(diag::UTF8Char, currentOffset() - 1);
+                sourceBuffer--;
+                addDiag(diag::UTF8Char, currentOffset());
 
-                // skip over UTF-8 sequences
-                int skip = utf8SeqBytes(c);
-                advance(std::min((int)(sourceEnd - sourceBuffer - 1), skip));
+                bool sawUTF8Error = false;
+                do {
+                    uint32_t unused;
+                    sawUTF8Error |= !scanUTF8Char(sawUTF8Error, &unused);
+                } while (!isASCII(peek()));
             }
             return create(TokenKind::Unknown);
     }
 }
 
 Token Lexer::lexStringLiteral() {
+    bool sawUTF8Error = false;
     SmallVectorSized<char, 128> stringBuffer;
     while (true) {
         size_t offset = currentOffset();
@@ -642,13 +647,27 @@ Token Lexer::lexStringLiteral() {
                     }
                     break;
                 default: {
-                    // '\%' is not an actual escape code but other tools silently allow it
-                    // and major UVM headers use it, so we'll issue a (fairly quiet) warning about
-                    // it. Otherwise issue a louder warning (on by default).
-                    DiagCode code =
-                        c == '%' ? diag::NonstandardEscapeCode : diag::UnknownEscapeCode;
-                    addDiag(code, offset) << c;
-                    stringBuffer.append(c);
+                    auto curr = --sourceBuffer;
+
+                    uint32_t unicodeChar;
+                    if (scanUTF8Char(sawUTF8Error, &unicodeChar)) {
+                        if (isPrintableUnicode(unicodeChar)) {
+                            // '\%' is not an actual escape code but other tools silently allow it
+                            // and major UVM headers use it, so we'll issue a (fairly quiet) warning
+                            // about it. Otherwise issue a louder warning (on by default).
+                            DiagCode code =
+                                c == '%' ? diag::NonstandardEscapeCode : diag::UnknownEscapeCode;
+                            addDiag(code, offset)
+                                << string_view(curr, (size_t)utf8Len((unsigned char)c));
+                        }
+                    }
+                    else {
+                        sawUTF8Error = true;
+                    }
+
+                    // Back up so that we handle this character as a normal char in the outer loop,
+                    // regardless of whether it's valid or not.
+                    sourceBuffer = curr;
                     break;
                 }
             }
@@ -672,9 +691,27 @@ Token Lexer::lexStringLiteral() {
             addDiag(diag::EmbeddedNull, offset);
             advance();
         }
-        else {
+        else if (isASCII(c)) {
             advance();
             stringBuffer.append(c);
+            sawUTF8Error = false;
+        }
+        else {
+            auto curr = sourceBuffer;
+
+            uint32_t unused;
+            sawUTF8Error |= !scanUTF8Char(sawUTF8Error, &unused);
+
+            // Regardless of whether the character sequence was valid or not
+            // we want to add the bytes to the string, to allow for cases where
+            // the source is actually something like latin-1 encoded. Ignoring the
+            // warning and carrying on will do the right thing for them.
+            int len = utf8Len((unsigned char)c);
+            if (len == 0)
+                len = 1;
+
+            for (int i = 0; i < len; i++)
+                stringBuffer.append(curr[i]);
         }
     }
 
@@ -696,7 +733,7 @@ Token Lexer::lexEscapeSequence(bool isMacroName) {
         return create(TokenKind::Unknown);
     }
 
-    while (isPrintable(c)) {
+    while (isPrintableASCII(c)) {
         advance();
         c = peek();
         if (isWhitespace(c))
@@ -907,8 +944,6 @@ Token Lexer::lexNumericLiteral() {
     // normal numeric literal
     SVInt intVal;
     if (!digits.empty()) {
-        static const double BitsPerDecimal = log2(10.0);
-
         double bitsDbl = ceil(BitsPerDecimal * double(digits.size()));
         bitwidth_t bits;
         if (bitsDbl <= double(SVInt::MAX_BITS))
@@ -1065,55 +1100,95 @@ void Lexer::scanWhitespace() {
 }
 
 void Lexer::scanLineComment() {
+    bool sawUTF8Error = false;
     while (true) {
         char c = peek();
-        if (isNewline(c))
-            break;
-
-        if (c == '\0') {
-            if (reallyAtEnd())
+        if (isASCII(c)) {
+            if (isNewline(c))
                 break;
 
-            // otherwise just error and ignore
-            errorCount++;
-            addDiag(diag::EmbeddedNull, currentOffset());
+            sawUTF8Error = false;
+            if (c == '\0') {
+                if (reallyAtEnd())
+                    break;
+
+                // otherwise just error and ignore
+                errorCount++;
+                addDiag(diag::EmbeddedNull, currentOffset());
+            }
+            advance();
+        }
+        else {
+            uint32_t unused;
+            sawUTF8Error |= !scanUTF8Char(sawUTF8Error, &unused);
         }
-        advance();
     }
     addTrivia(TriviaKind::LineComment);
 }
 
 void Lexer::scanBlockComment() {
+    bool sawUTF8Error = false;
     while (true) {
         char c = peek();
-        if (c == '\0') {
-            if (reallyAtEnd()) {
-                addDiag(diag::UnterminatedBlockComment, currentOffset());
+        if (isASCII(c)) {
+            sawUTF8Error = false;
+            if (c == '*' && peek(1) == '/') {
+                advance(2);
                 break;
             }
+            else if (c == '/' && peek(1) == '*') {
+                // nested block comments disallowed by the standard; ignore and continue
+                addDiag(diag::NestedBlockComment, currentOffset());
+                advance(2);
+            }
+            else if (c == '\0') {
+                if (reallyAtEnd()) {
+                    addDiag(diag::UnterminatedBlockComment, currentOffset());
+                    break;
+                }
 
-            // otherwise just error and ignore
-            errorCount++;
-            addDiag(diag::EmbeddedNull, currentOffset());
-            advance();
-        }
-        else if (c == '*' && peek(1) == '/') {
-            advance(2);
-            break;
-        }
-        else if (c == '/' && peek(1) == '*') {
-            // nested block comments disallowed by the standard; ignore and continue
-            addDiag(diag::NestedBlockComment, currentOffset());
-            advance(2);
+                // otherwise just error and ignore
+                errorCount++;
+                addDiag(diag::EmbeddedNull, currentOffset());
+                advance();
+            }
+            else {
+                advance();
+            }
         }
         else {
-            advance();
+            uint32_t unused;
+            sawUTF8Error |= !scanUTF8Char(sawUTF8Error, &unused);
         }
     }
 
     addTrivia(TriviaKind::BlockComment);
 }
 
+bool Lexer::scanUTF8Char(bool alreadyErrored, uint32_t* code) {
+    int error;
+    auto curr = sourceBuffer;
+    if (sourceBuffer + 4 < sourceEnd) {
+        sourceBuffer = utf8Decode(sourceBuffer, code, &error);
+    }
+    else {
+        char buf[4] = {};
+        memcpy(buf, sourceBuffer, size_t(sourceEnd - sourceBuffer - 1));
+
+        auto next = utf8Decode(buf, code, &error);
+        sourceBuffer += next - buf;
+    }
+
+    if (error) {
+        errorCount++;
+        if (!alreadyErrored)
+            addDiag(diag::InvalidUTF8Seq, (size_t)(curr - originalBegin));
+        return false;
+    }
+
+    return true;
+}
+
 template<typename... Args>
 Token Lexer::create(TokenKind kind, Args&&... args) {
     SourceLocation location(bufferId, size_t(marker - originalBegin));
