@@ -450,8 +450,13 @@ const Expression* ClassType::getBaseConstructorCall() const {
     if (!classSyntax.extendsClause)
         return nullptr;
 
-    // If we have a constructor, find whether it invokes super.new in its body.
     ensureElaborated();
+
+    ASSERT(baseClass);
+    if (baseClass->isError())
+        return nullptr;
+
+    // If we have a constructor, find whether it invokes super.new in its body.
     if (auto ourConstructor = find("new")) {
         auto checkForSuperNew = [&](const Statement& stmt) {
             if (stmt.kind == StatementKind::ExpressionStatement) {
@@ -922,6 +927,9 @@ ConstraintBlockSymbol* ConstraintBlockSymbol::fromSyntax(
         return nullptr;
     }
 
+    if (scope.asSymbol().kind != SymbolKind::ClassType)
+        scope.addDiag(diag::ConstraintNotInClass, syntax.sourceRange());
+
     auto nameToken = syntax.name->getLastToken();
     auto result =
         comp.emplace<ConstraintBlockSymbol>(comp, nameToken.valueText(), nameToken.location());
@@ -940,7 +948,7 @@ ConstraintBlockSymbol* ConstraintBlockSymbol::fromSyntax(
         }
     }
 
-    if (!result->isStatic)
+    if (!result->isStatic && scope.asSymbol().kind == SymbolKind::ClassType)
         result->addThisVar(scope.asSymbol().as<ClassType>());
 
     return result;
@@ -972,14 +980,14 @@ ConstraintBlockSymbol& ConstraintBlockSymbol::fromSyntax(const Scope& scope,
         }
     }
 
-    if (result->isPure && scope.asSymbol().kind == SymbolKind::ClassType) {
+    if (scope.asSymbol().kind == SymbolKind::ClassType) {
         auto& classType = scope.asSymbol().as<ClassType>();
-        if (!classType.isAbstract)
+        if (result->isPure && !classType.isAbstract)
             scope.addDiag(diag::PureConstraintInAbstract, nameToken.range());
-    }
 
-    if (!result->isStatic)
-        result->addThisVar(scope.asSymbol().as<ClassType>());
+        if (!result->isStatic)
+            result->addThisVar(classType);
+    }
 
     return *result;
 }
@@ -1000,8 +1008,8 @@ const Constraint& ConstraintBlockSymbol::getConstraints() const {
         auto& comp = outerScope.getCompilation();
 
         auto [declSyntax, index, used] = comp.findOutOfBlockDecl(outerScope, parentSym.name, name);
-        if (!declSyntax || declSyntax->kind != SyntaxKind::ConstraintDeclaration) {
-            if (!isPure) {
+        if (!declSyntax || declSyntax->kind != SyntaxKind::ConstraintDeclaration || name.empty()) {
+            if (!isPure && !name.empty()) {
                 DiagCode code = isExplicitExtern ? diag::NoMemberImplFound : diag::NoConstraintBody;
                 outerScope.addDiag(code, location) << name;
             }
