@@ -129,14 +129,14 @@ string_view SourceManager::getMacroName(SourceLocation location) const {
 }
 
 bool SourceManager::isFileLoc(SourceLocation location) const {
-    if (location == SourceLocation::NoLocation)
+    if (location.buffer() == SourceLocation::NoLocation.buffer())
         return false;
 
     return getFileInfo(location.buffer());
 }
 
 bool SourceManager::isMacroLoc(SourceLocation location) const {
-    if (location == SourceLocation::NoLocation)
+    if (location.buffer() == SourceLocation::NoLocation.buffer())
         return false;
 
     auto buffer = location.buffer();
@@ -419,6 +419,13 @@ void SourceManager::addDiagnosticDirective(SourceLocation location, string_view
     }
 }
 
+span<const SourceManager::DiagnosticDirectiveInfo> SourceManager::getDiagnosticDirectives(
+    BufferID buffer) const {
+    if (auto it = diagDirectives.find(buffer); it != diagDirectives.end())
+        return it->second;
+    return {};
+}
+
 SourceManager::FileInfo* SourceManager::getFileInfo(BufferID buffer) {
     if (!buffer)
         return nullptr;
